package mpi;
import mpi.*;

public class Comm {
  public final static int NULL  = 0;
  public final static int SELF  = 1;
  public final static int WORLD = 2;

  private static native void init();

  public Comm() {}
  public Comm(int Type) { GetComm(Type);}
  private Comm(long _handle) { handle = _handle;}  
  protected native void GetComm(int Type);

  public native int Size();
  public native int Rank();

  public Comm Dup()
  {  return new Comm(dup());
     }
  public native long dup();

  public static native int Compare(Comm comm1, Comm comm2);

//  public native void Free();
  public native void finalize();

  public native long group();
  public Group Group()
  { return new Group(group());
	}

  // Inter-communication

  public native boolean Test_inter();

  public Intercomm Create_intercomm(Comm local_comm, int local_leader,
	int remote_leader, int tag)
  { return new Intercomm(GetIntercomm(local_comm,local_leader,remote_leader,tag)); }
  public native long GetIntercomm(Comm local_comm, int local_leader,
	int remote_leader, int tag);

  // Blocking Send and Recv

  public native void Send(Object buf, int offset, int baseSize,
    int count, Datatype type, int dest, int tag);
  public void Send(Object buf, int offset, int count,
    Datatype type, int dest, int tag)
  { Send(buf, offset, type.baseSize(), count,
      type, dest, tag);}

  private native Status Recv(Object buf, int offset,
    int count, Datatype type, int source, int tag, Status stat);
  public Status Recv(Object buf, int offset, int count, 
    Datatype type, int source, int tag)
  { return Recv(buf, offset*type.baseSize(), 
      count, type, source, tag, new Status());}
  
  // Send-Recv

  private native Status Sendrecv(
    Object sbuf, int soffset, int scount,
    Datatype stype, int dest, int stag,
    Object rbuf, int roffset, int rcount, 
    Datatype rtype, int source, int rtag, Status stat);
  public Status Sendrecv(
    Object sbuf, int soffset, int scount,
    Datatype stype, int dest, int stag,
    Object rbuf, int roffset, int rcount,
    Datatype rtype, int source, int rtag)
  { return Sendrecv(sbuf, soffset*stype.baseSize(),
      scount, stype, dest, stag,
      rbuf, roffset*rtype.baseSize(),
      rcount, rtype, source, rtag,
      new Status());}

  private native Status Sendrecv_replace(
    Object buf, int offset, int count,
    Datatype type, int dest, int stag,
    int source, int rtag, Status stat);
  public Status Sendrecv_replace(
    Object buf, int offset, int count,
    Datatype type, int dest, int stag,
    int source, int rtag)
  { return Sendrecv_replace(buf, offset*type.baseSize(),
      count, type, dest, stag,
      source, rtag,
      new Status());}

  // Communication Modes

  public native void bsend(Object buf, int offset, int count,Datatype type, int dest, int tag);
  public void Bsend(Object buf, int offset, int count,Datatype type, int dest, int tag)
{  bsend(buf,offset*type.baseSize(), count, type, dest, tag);
}

  public native void ssend(Object buf, int offset, int count,Datatype type, int dest, int tag);
 public void Ssend(Object buf, int offset, int count,Datatype type, int dest, int tag){
ssend(buf, offset*type.baseSize(), count,type, dest, tag);
}

  public native void rsend(Object buf, int offset, int count,Datatype type, int dest, int tag);
 public void Rsend(Object buf, int offset, int count,Datatype type, int dest, int tag)
{rsend(buf, offset*type.baseSize(), count, type, dest, tag);
}

  // Nonblocking communication 
  public native Request Isend(Object buf, int offset, int count,Datatype type, int dest, int tag, Request req);
public Request Isend(Object buf, int offset, int count,Datatype type, int dest, int tag)
{   return Isend(buf, offset*type.baseSize(), count, type, dest, tag, new Request());
}

  public native Request Ibsend(Object buf, int offset, int count,Datatype type, int dest, int tag, Request req);
 public Request Ibsend(Object buf, int offset, int count,Datatype type, int dest, int tag)
{ return Ibsend(buf, offset*type.baseSize(), count, type, dest, tag, 
new Request());
}

  public native Request Issend(Object buf, int offset, int count,Datatype type, int dest, int tag, Request req);
public Request Issend(Object buf, int offset, int count,Datatype type, int dest, int tag)
{return Issend(buf, offset*type.baseSize(), count, type, dest, tag, 
new Request());
}

  public native Request Irsend(Object buf, int offset, int count,Datatype type, int dest, int tag,Request req);
 public Request Irsend(Object buf, int offset, int count,Datatype type, int dest, int tag)
{return Irsend(buf, offset*type.baseSize(), count, type, dest, tag, 
new Request()); 
}

  public native Request Irecv(Object buf, int offset, int count,Datatype type, int dest, int tag,Request req);
 public Request Irecv(Object buf, int offset, int count,Datatype type, int dest, int tag)
{return Irecv(buf, offset*type.baseSize(), count, type, dest, tag, 
new Request());
}



  // Persistent communication  requests
  public native Prequest Send_init(Object buf, int offset, int count,Datatype type, int dest, int tag,Prequest req);
public Prequest Send_init(Object buf, int offset, int count,Datatype type, int dest, int tag)
{return Send_init(buf, offset*type.baseSize(), count, type, dest, tag, 
new Prequest());
}

  public native Prequest Bsend_init(Object buf, int offset, int count,Datatype type, int dest, int tag,Prequest req);
 public Prequest Bsend_init(Object buf, int offset, int count,Datatype type, int dest, int tag)
{return Bsend_init(buf, offset*type.baseSize(), count, type, dest, tag, 
new Prequest());
}

  public native Prequest Ssend_init(Object buf, int offset, int count,Datatype type, int dest, int tag,Prequest req);
public Prequest Ssend_init(Object buf, int offset, int count,Datatype type, int dest, int tag)
{return Ssend_init(buf, offset*type.baseSize(), count, type, dest, tag, 
new Prequest());
}

  public native Prequest Rsend_init(Object buf, int offset, int count,Datatype type, int dest, int tag,Prequest req);
public Prequest Rsend_init(Object buf, int offset, int count,Datatype type, int dest, int tag)
{
return Rsend_init(buf, offset*type.baseSize(), count, type, dest, tag, 
new Prequest());
}

  public native Prequest Recv_init(Object buf, int offset, int count,Datatype type, int dest, int tag,Prequest req);
  public Prequest Recv_init(Object buf, int offset, int count,Datatype type, int dest, int tag)
{
return Recv_init(buf, offset*type.baseSize(), count, type, dest, tag, 
new Prequest());
}

  // Pack and Unpack 
  public native int pack(Object inbuf, int offset, int incount,Datatype type, 
byte[] outbuf, int position);
public int Pack(Object inbuf, int offset, int incount,Datatype type, 
byte[] outbuf, int position)
{return pack(inbuf, offset*type.baseSize(), incount, type, 
outbuf, position);
}


  public native int unpack(byte[] inbuf, int position, Object outbuf, 
int offset, int outcount, Datatype type);
 public int Unpack(byte[] inbuf, int position, Object outbuf, 
int offset, int outcount, Datatype type)
{return unpack(inbuf, position, outbuf,
 offset*type.baseSize(), outcount, type);
}


  public native int Pack_size(int incount, Datatype type);

  // Probe and Cancel
  public native Status Iprobe(int source, int tag,Status stat);
  public native Status Probe(int source, int tag,Status stat);

  public Status Iprobe(int source, int tag)
  { return Iprobe(source,tag,new Status()); }
  public Status Probe(int source, int tag)
  { return Probe(source,tag,new Status()); }

  // Caching
  //public native void Attr_put(int keyval, Object attribute_val);
  //public native Object Attr_get(int keyval);
  public native void Attr_put(int keyval, int attribute_val);
  public native int Attr_get(int keyval);
  public native void Attr_delete(int keyval);

  // Process Topologies
  public native int Topo_test();

  // Enviromental Management 
  public native void Abort(int errorcode);
  // Error handle
  static public native void Errhandler_set(Errhandler errhandler);
  static private native long errorhandler_get();
  static public Errhandler Errorhandler_get()
  { return new Errhandler(errorhandler_get());
    }                           

  protected long handle;

  static {
    init();
  }

}

