package mpi;
import mpi.*;

public class MPI {

  static int MAX_PROCESSOR_NAME = 256;

//  static public Comm COMM_WORLD;
  static public Intracomm COMM_WORLD;
  static public Comm COMM_SELF;

  static public int GRAPH, CART;
  static public int ANY_SOURCE, ANY_TAG;

  static public Op MAX, MIN, SUM, PROD, LAND, BAND,
    LOR, BOR, LXOR, BXOR, MINLOC, MAXLOC;
  static public Datatype BYTE, CHAR, SHORT, BOOLEAN, 
    INT, LONG, FLOAT, DOUBLE, PACKED, LB, UB;
  static public Datatype SHORT2, INT2, LONG2, FLOAT2, DOUBLE2;

  static public Request REQUEST_NULL;
  static public Group GROUP_EMPTY;

  static public int PROC_NULL;
  static public int BSEND_OVERHEAD;
  static public int UNDEFINED;
  static public int IDENT, CONGRUENT, SIMILAR, UNEQUAL;
  static public int TAG_UB, HOST, IO;

  static public Call_back_null CALL_BACK_NULL;
  static public Call_back_dup  CALL_BACK_DUP;

  static Errhandler ERRORS_ARE_FATAL, ERRORS_RETURN;

  static private native void InitNative(String[] args);
  static private native void SetConstant();
  static public void Init(String[] args) {
    InitNative(args);
    MAX = new Op(Op.MAX);
    MIN = new Op(Op.MIN);
    SUM = new Op(Op.SUM);
    PROD = new Op(Op.PROD);
    LAND = new Op(Op.LAND);
    BAND = new Op(Op.BAND);
    LOR = new Op(Op.LOR);
    BOR = new Op(Op.BOR);
    LXOR = new Op(Op.LXOR);
    BXOR = new Op(Op.BXOR);
    MINLOC = new Op(Op.MINLOC);
    MAXLOC = new Op(Op.MAXLOC);
    BYTE = new Datatype(Datatype.BYTE);
    CHAR = new Datatype(Datatype.CHAR);
    SHORT = new Datatype(Datatype.SHORT);
    BOOLEAN = new Datatype(Datatype.BOOLEAN);
    INT = new Datatype(Datatype.INT);
    LONG = new Datatype(Datatype.LONG);
    FLOAT = new Datatype(Datatype.FLOAT);
    DOUBLE = new Datatype(Datatype.DOUBLE);
    PACKED = new Datatype(Datatype.PACKED);
    LB = new Datatype(Datatype.LB);
    UB = new Datatype(Datatype.UB);

	GROUP_EMPTY = new Group(Group.NULL);	
	REQUEST_NULL = new Request(Request.NULL);
    
    // Constant
    SetConstant(); 

    CALL_BACK_NULL = new Call_back_null();
    CALL_BACK_DUP = new Call_back_dup();
    
    ERRORS_ARE_FATAL= new Errhandler(Errhandler.FATAL);
    ERRORS_RETURN = new Errhandler(Errhandler.RETURN);       

//    COMM_WORLD = new Comm(Comm.WORLD);
    COMM_WORLD = new Intracomm(Intracomm.WORLD);
  }

  static public native void Finalize();

  static public native double Wtime();
  static public native double Wtick();

  static private native int Get_processor_name(byte[] buf);
  static public String Get_processor_name() {
    byte[] buf = new byte[MAX_PROCESSOR_NAME];
	int lengh=Get_processor_name(buf);
    return new String(buf,0,0,lengh);
  }

  static public native boolean Initialized();

  // Buffer allocation
  static public native void Buffer_attach(byte[] buf);
  static public native void Buffer_detach(byte[] buf);
   
  // Caching ...
  static public native int Keyval_create(Call_back call_back);
  static public native void Keyval_free(int keyval);

  static {
    System.loadLibrary("mpi");
  }
}

class Call_back_null extends Call_back {
    public Object copy(Comm oldcomm, int keyval,
                    Object attribute_val_in)
	{ return null; }
    public void delete(Comm comm, int keyval,
                    Object attribute_val)
	{ }
}

class Call_back_dup extends Call_back {
    public Object copy(Comm oldcomm, int keyval,
                    Object attribute_val_in)
    { return attribute_val_in; }
    public void delete(Comm comm, int keyval,
                    Object attribute_val)
    { }
}
