#include "mpi.h"
#include "mpi_Cartcomm.h"

extern jfieldID CommhandleID;

/*
 * Class:     mpi_Cartcomm
 * Method:    Get
 * Signature: (I)Lmpi/CartParms;
 */
JNIEXPORT jobject JNICALL Java_mpi_Cartcomm_Get
  (JNIEnv *env, jobject jthis)
{
  jintArray dims, coords;
  jbooleanArray periods;
  jint *ds, *cs; 
  jboolean *ps;
  jboolean isCopy=JNI_TRUE;
  int maxdims;

  jclass cartparms_class=(*env)->FindClass(env,"mpi/CartParms");
  jfieldID dimsID,periodsID,coordsID;
  jmethodID handleConstructorID = (*env)->GetMethodID(env,
	 cartparms_class, "<init>", "()V");
  jobject cartparms=(*env)->NewObject(env,cartparms_class, handleConstructorID);

  MPI_Cartdim_get((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)),&maxdims);
  dims=(*env)->NewIntArray(env,maxdims);
  periods=(*env)->NewBooleanArray(env,maxdims);
  coords=(*env)->NewIntArray(env,maxdims);
  ds=(*env)->GetIntArrayElements(env,dims,&isCopy);
  ps=(*env)->GetBooleanArrayElements(env,periods,&isCopy);
  cs=(*env)->GetIntArrayElements(env,coords,&isCopy);

  MPI_Cart_get((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)), 
	maxdims, (int*)ds, (int*)ps, (int*)cs);

  (*env)->ReleaseIntArrayElements(env,dims,ds,0);
  (*env)->ReleaseBooleanArrayElements(env,periods,ps,0);
  (*env)->ReleaseIntArrayElements(env,coords,cs,0);

  dimsID=(*env)->GetFieldID(env,cartparms_class,"dims","[I");
  periodsID=(*env)->GetFieldID(env,cartparms_class,"periods","[Z");
  coordsID=(*env)->GetFieldID(env,cartparms_class , "coords", "[I");

  (*env)->SetObjectField(env, cartparms, dimsID, dims);
  (*env)->SetObjectField(env, cartparms, periodsID, periods);
  (*env)->SetObjectField(env, cartparms, coordsID, coords);
  
  /* printf("Get finished.\n"); */
  return cartparms;
}

/*
 * Class:     mpi_Cartcomm
 * Method:    Shift
 * Signature: (II)Lmpi/ShiftParms;
 */
JNIEXPORT jobject JNICALL Java_mpi_Cartcomm_Shift
  (JNIEnv *env, jobject jthis, jint direction, jint disp)
{
  int sr, dr;
  jclass shiftparms_class=(*env)->FindClass(env,"mpi/ShiftParms");
  jfieldID rsID,rdID;
  jmethodID handleConstructorID = (*env)->GetMethodID(env,
     shiftparms_class, "<init>", "()V");
  jobject shiftparms=(*env)->NewObject(env,shiftparms_class,
handleConstructorID);

  MPI_Cart_shift((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)), 
	direction, disp, &sr, &dr);
  rsID=(*env)->GetFieldID(env,shiftparms_class,"rank_source","I");
  rdID=(*env)->GetFieldID(env,shiftparms_class,"rank_dest", "I");
  (*env)->SetIntField(env, shiftparms, rsID, sr);
  (*env)->SetIntField(env, shiftparms, rdID, dr);
  /* printf("Shift finished.\n"); */
  return shiftparms;
}

/*
 * Class:     mpi_Cartcomm
 * Method:    Coords
 * Signature: (I)[I
 */
JNIEXPORT jintArray JNICALL Java_mpi_Cartcomm_Coords
  (JNIEnv *env, jobject jthis, jint rank)
{
	jint *coords;
    jboolean isCopy=JNI_TRUE;
    jintArray jcoords;
    int maxdims;
/*
	jclass jthis_class=(*env)->FindClass(env,"mpi/Cartcomm");
    jfieldID maxdimsID=(*env)->GetFieldID(env,jthis_class,"maxdims","I");
    maxdims=(*env)->GetIntField(env,jthis, maxdimsID);
*/
    MPI_Cartdim_get((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)),
		&maxdims);
	jcoords=(*env)->NewIntArray(env,maxdims);
	coords=(*env)->GetIntArrayElements(env,jcoords,&isCopy);
	MPI_Cart_coords((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)),
    	rank,maxdims,(int*)coords);
	(*env)->ReleaseIntArrayElements(env,jcoords,coords,0);
	return jcoords;
}

/*
 * Class:     mpi_Cartcomm
 * Method:    Map
 * Signature: ([I[Z)I
 */
JNIEXPORT jint JNICALL Java_mpi_Cartcomm_Map
  (JNIEnv *env, jobject jthis, jintArray dims, jbooleanArray periods)
{
    int newrank;
    jint *ds;
    jboolean *ps;
    jboolean isCopy=JNI_TRUE;
    int ndims;

    ndims=(*env)->GetArrayLength(env,dims);
    ds=(*env)->GetIntArrayElements(env,dims,&isCopy);
    ps=(*env)->GetBooleanArrayElements(env,periods,&isCopy);

    MPI_Cart_map((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)),
        ndims,(int*)ds,(int*)ps, &newrank);
    (*env)->ReleaseIntArrayElements(env,dims,ds,0);
    (*env)->ReleaseBooleanArrayElements(env,periods,ps,0);
    return newrank;
}

/*
 * Class:     mpi_Cartcomm
 * Method:    Rank
 * Signature: ([I)I
 */
JNIEXPORT jint JNICALL Java_mpi_Cartcomm_Rank
  (JNIEnv *env, jobject jthis, jintArray coords)
{
	int rank;
	jint *crds;
	jboolean isCopy=JNI_TRUE;
	crds=(*env)->GetIntArrayElements(env,coords,&isCopy);
	MPI_Cart_rank((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)),
        (int*)crds, &rank);
	(*env)->ReleaseIntArrayElements(env,coords,crds,0);
	return rank;
}

/*
 * Class:     mpi_Cartcomm
 * Method:    Sub
 * Signature: ([Z)Lmpi/Cartcomm;
 */
JNIEXPORT jlong JNICALL Java_mpi_Cartcomm_sub
  (JNIEnv *env, jobject jthis, jbooleanArray remain_dims)
{
	MPI_Comm newcomm;
	jboolean *re_ds;
	jboolean isCopy=JNI_TRUE;

	re_ds=(*env)->GetBooleanArrayElements(env,remain_dims,&isCopy);
	MPI_Cart_sub((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)),
        (int*)re_ds, &newcomm);
	(*env)->ReleaseBooleanArrayElements(env,remain_dims,re_ds,0);

	return (jlong)newcomm;
}

/*
 * Class:     mpi_Cartcomm
 * Method:    Dims_create
 * Signature: (II[I)V
 */
JNIEXPORT void JNICALL Java_mpi_Cartcomm_Dims_1create
  (JNIEnv *env, jclass jthis, jint nnodes, jint ndims,jintArray dims )
{
	jint *cdims;
	jboolean isCopy=JNI_TRUE;
	cdims=(*env)->GetIntArrayElements(env,dims,&isCopy);
        MPI_Dims_create(nnodes,ndims,(int*)cdims);
	(*env)->ReleaseIntArrayElements(env,dims,cdims,0);
}
