#include <mpi.h>
#include "mpi_Datatype.h"

jfieldID DatatypehandleID;

/*
 * Class:     mpi_Datatype
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Datatype_init
  (JNIEnv *env, jclass thisClass)
{
  DatatypehandleID = (*env)->GetFieldID(env,thisClass,"handle","J");
}

/*
 * Class:     mpi_Datatype
 * Method:    GetDatatype
 * Signature: (I)J
 */
JNIEXPORT void JNICALL Java_mpi_Datatype_GetDatatype
  (JNIEnv *env, jobject jthis, jint type) 
{
  static MPI_Datatype Dts[] = { MPI_DATATYPE_NULL, MPI_CHAR, MPI_SHORT, MPI_SHORT,
    MPI_LONG, MPI_LONG, MPI_LONG_LONG_INT, MPI_FLOAT, MPI_DOUBLE,
    MPI_PACKED, MPI_LB, MPI_UB};
  (*env)->SetLongField(env,jthis, DatatypehandleID, (jlong)Dts[type]);
}

/*
 * Class:     mpi_Datatype
 * Method:    Size
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_mpi_Datatype_Size
  (JNIEnv *env, jobject jthis)
{
  int result;
  MPI_Type_size((MPI_Datatype)((*env)->GetLongField(env,jthis,DatatypehandleID)), &result);
  return result;
}

/*
 * Class:     mpi_Datatype
 * Method:    Extent
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_mpi_Datatype_Extent
  (JNIEnv *env, jobject jthis)
{
  MPI_Aint  result;
  MPI_Type_extent((MPI_Datatype)((*env)->GetLongField(env,jthis,DatatypehandleID)),
&result);
  return result;
}

/*
 * Class:     mpi_Datatype
 * Method:    Lb
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_mpi_Datatype_Lb
  (JNIEnv *env, jobject jthis)
{
  MPI_Aint  result;
  MPI_Type_lb((MPI_Datatype)((*env)->GetLongField(env,jthis,DatatypehandleID)),
&result);
  return result;
}

/*
 * Class:     mpi_Datatype
 * Method:    Ub
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_mpi_Datatype_Ub
  (JNIEnv *env, jobject jthis)
{
  MPI_Aint  result;
  MPI_Type_ub((MPI_Datatype)((*env)->GetLongField(env,jthis,DatatypehandleID)),
&result);
  return result;
}

/*
 * Class:     mpi_Datatype
 * Method:    Commit
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Datatype_Commit
  (JNIEnv *env, jobject jthis)
{
  MPI_Datatype type;
  type=(MPI_Datatype)((*env)->GetLongField(env,jthis,DatatypehandleID));
  MPI_Type_commit(&type);
}

/*
 * Class:     mpi_Datatype
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Datatype_finalize
  (JNIEnv *env, jobject jthis)
{
  MPI_Datatype type;
  type=(MPI_Datatype)((*env)->GetLongField(env,jthis,DatatypehandleID));
  MPI_Type_free(&type);
  (*env)->SetLongField(env,jthis, DatatypehandleID,0);
}

/*
 * Class:     mpi_Datatype
 * Method:    GetContiguous
 * Signature: (I)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Datatype_GetContiguous
  (JNIEnv *env, jobject jthis, jint count,jobject oldtype)
{
  MPI_Datatype type;
  MPI_Type_contiguous(count, (MPI_Datatype)((*env)->GetLongField(env,oldtype,DatatypehandleID)), &type);
  return (jlong)type; 
}

/*
 * Class:     mpi_Datatype
 * Method:    GetVector
 * Signature: (III)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Datatype_GetVector
  (JNIEnv *env, jobject jthis, jint count, jint blocklength, jint stride, jobject oldtype)
{
  MPI_Datatype type;
  MPI_Type_vector(count, blocklength, stride, 
	(MPI_Datatype)((*env)->GetLongField(env,oldtype,DatatypehandleID)), &type);
  return (jlong)type;
}

/*
 * Class:     mpi_Datatype
 * Method:    GetHvector
 * Signature: (III)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Datatype_GetHvector
  (JNIEnv *env, jobject jthis, jint count, jint blocklength, jint stride,jobject oldtype)
{
  MPI_Datatype type;
  MPI_Type_hvector(count, blocklength, stride, 
    (MPI_Datatype)((*env)->GetLongField(env,oldtype,DatatypehandleID)), &type);
  return (jlong)type;
}

/*
 * Class:     mpi_Datatype
 * Method:    GetIndexed
 * Signature: (I[I[I)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Datatype_GetIndexed
  (JNIEnv *env, jobject jthis, jintArray blocklengths, jintArray
displacements, jobject oldtype)
{
  MPI_Datatype type;
  int count=(*env)->GetArrayLength(env,blocklengths);
  jboolean isCopy=JNI_TRUE;
  jint *lengths; jint *disps;
  lengths=(*env)->GetIntArrayElements(env,blocklengths,&isCopy);
  disps = (*env)->GetIntArrayElements(env,displacements,&isCopy);
  MPI_Type_indexed(count, (int*)lengths, (int*)disps, 
    (MPI_Datatype)((*env)->GetLongField(env,oldtype,DatatypehandleID)), &type);
  (*env)->ReleaseIntArrayElements(env,blocklengths,lengths,0);
  (*env)->ReleaseIntArrayElements(env,displacements,disps,0);
  return (jlong)type;
}

/*
 * Class:     mpi_Datatype
 * Method:    GetHindexed
 * Signature: (I[I[I)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Datatype_GetHindexed
  (JNIEnv *env, jobject jthis, jintArray blocklengths, jintArray
displacements,jobject oldtype)
{
  MPI_Datatype type;
  int count=(*env)->GetArrayLength(env,blocklengths);
  jboolean isCopy=JNI_TRUE;
  jint *lengths; jint *disps;
  lengths=(*env)->GetIntArrayElements(env,blocklengths,&isCopy);
  disps = (*env)->GetIntArrayElements(env,displacements,&isCopy);
  MPI_Type_hindexed(count, (int*)lengths, (MPI_Aint*)disps,
    (MPI_Datatype)((*env)->GetLongField(env,oldtype,DatatypehandleID)), &type);
  (*env)->ReleaseIntArrayElements(env,blocklengths,lengths,0);
  (*env)->ReleaseIntArrayElements(env,displacements,disps,0);
  return (jlong)type;
}

/*
 * Class:     mpi_Datatype
 * Method:    GetStruct
 * Signature: (I[I[I[Lmpi/Datatype;)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Datatype_GetStruct
  (JNIEnv *env, jobject jthis, jintArray blocklengths, jintArray
displacements,jobjectArray datatypes)
{
  MPI_Datatype type;
  int i;
  int count=(*env)->GetArrayLength(env,blocklengths);
  MPI_Datatype *types=(MPI_Datatype*)calloc(count, sizeof(MPI_Datatype));
  jboolean isCopy=JNI_TRUE;
  jint *lengths; jint *disps;
  lengths=(*env)->GetIntArrayElements(env,blocklengths,&isCopy);
  disps = (*env)->GetIntArrayElements(env,displacements,&isCopy);
  for(i=0; i<count; i++) {
	types[i]=(MPI_Datatype)((*env)->GetLongField(env,
		(*env)->GetObjectArrayElement(env,datatypes,i),DatatypehandleID));
  } 
  MPI_Type_struct(count, (int*)lengths, (MPI_Aint*)disps, types, &type);
  (*env)->ReleaseIntArrayElements(env,blocklengths,lengths,0);
  (*env)->ReleaseIntArrayElements(env,displacements,disps,0);
  free(types);
  return (jlong)type;
}
