#include <mpi.h>
#include "mpi_Comm.h"

#include "mpi_Intracomm.h"

extern jfieldID CommhandleID, DatatypehandleID, OphandleID;
extern jfieldID GrouphandleID;

jbyte *getBufPtr(JNIEnv *env,jobject buf, jobject type);
void releaseBufPtr(JNIEnv *env,jobject buf,jobject type, jbyte* bufptr);

/*
 * Class:     mpi_Intracomm
 * Method:    split
 * Signature: (II)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Intracomm_split
  (JNIEnv *env, jobject jthis, jint colour, jint key)
{
    MPI_Comm newcomm;

MPI_Comm_split((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)), colour, key,&newcomm);
    return (jlong)newcomm;
}

/*
 * Class:     mpi_Intracomm
 * Method:    creat
 * Signature: (Lmpi/Group;)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Intracomm_creat
  (JNIEnv *env, jobject jthis, jobject group)
{
    MPI_Comm newcomm;

MPI_Comm_create((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)),
        (MPI_Group)((*env)->GetLongField(env,group,GrouphandleID)), &newcomm);
    if(newcomm==MPI_COMM_NULL)
        return NULL;
    return (jlong)newcomm;
}

/*
 * Class:     mpi_Intracomm
 * Method:    Barrier
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_Barrier
  (JNIEnv *env, jobject jthis)
{
	MPI_Barrier((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
}

/*
 * Class:     mpi_Intracomm
 * Method:    GetCart
 * Signature: ([I[ZZ)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Intracomm_GetCart
  (JNIEnv *env, jobject jthis, jintArray dims, jbooleanArray periods, 
	jboolean reorder)
{
  MPI_Comm cart;
  int ndims=(*env)->GetArrayLength(env,dims);
  jboolean isCopy=JNI_TRUE;
  jint *ds; jboolean *ps;
  ds=(*env)->GetIntArrayElements(env,dims,&isCopy);
  ps=(*env)->GetBooleanArrayElements(env,periods,&isCopy);
  MPI_Cart_create((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)), 
	ndims, (int*)ds, (int*)ps, reorder, &cart);
  (*env)->ReleaseIntArrayElements(env,dims,ds,0);
  (*env)->ReleaseBooleanArrayElements(env,periods,ps,0);
  /* printf("Create cart\n"); */
  return (jlong)cart;
}

/*
 * Class:     mpi_Intracomm
 * Method:    GetGraph
 * Signature: ([I[IZ)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Intracomm_GetGraph
  (JNIEnv *env, jobject jthis, jintArray index, jintArray edges, jboolean reorder)
{
  MPI_Comm graph;
  int nnodes=(*env)->GetArrayLength(env,index);
  jboolean isCopy=JNI_TRUE;
  jint *ind, *edg;
  ind=(*env)->GetIntArrayElements(env,index,&isCopy);
  edg=(*env)->GetIntArrayElements(env,edges,&isCopy);
  MPI_Graph_create((MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)),
    nnodes, (int*)ind, (int*)edg, reorder, &graph);
  (*env)->ReleaseIntArrayElements(env,index,ind,0);
  (*env)->ReleaseIntArrayElements(env,edges,edg,0);
  /* printf("Create Graph\n"); */
  return (jlong)graph;
}

/*
 * Class:     mpi_Intracomm
 * Method:    bcast
 * Signature: (Ljava/lang/Object;IILmpi/Datatype;I)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_bcast
  (JNIEnv *env, jobject jthis, jobject buf, jint offset, jint count, jobject
type, jint root)
{
   jbyte * bufptr=getBufPtr(env,buf,type);
   MPI_Bcast(bufptr+offset,count,(MPI_Datatype)((*env)->GetLongField(env,type,DatatypehandleID)),root,(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   releaseBufPtr(env,buf,type,bufptr);
}


/*
 * Class:     mpi_Intracomm
 * Method:    Gather
 * Signature:
(Ljava/lang/Object;IILmpi/Datatype;Ljava/lang/Object;IILmpi/Dataty
pe;I)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_gather
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, jint
sendcount, jobject sendtype, jobject recvbuf, jint recvoffset, jint recvcount,
jobject recvtype, jint root)
{
   jbyte * sendptr=getBufPtr(env,sendbuf,sendtype);
   jbyte * recvptr=getBufPtr(env,recvbuf,recvtype);
   MPI_Gather(sendptr+sendoffset, sendcount,
	(MPI_Datatype)((*env)->GetLongField(env,sendtype,DatatypehandleID)), 
	recvptr+recvoffset, recvcount,
	(MPI_Datatype)((*env)->GetLongField(env,recvtype,DatatypehandleID)),
	root,(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   releaseBufPtr(env,sendbuf,sendtype,sendptr);
   releaseBufPtr(env,recvbuf,recvtype,recvptr);
}

/*
 * Class:     mpi_Intracomm
 * Method:    Gatherv
 * Signature:
(Ljava/lang/Object;IILmpi/Datatype;Ljava/lang/Object;I[I[ILmpi/Dat
atype;I)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_gatherv
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, jint sendcount,jobject sendtype, jobject recvbuf, jint recvoffset,
jintArray recvcount, jintArray displs, jobject recvtype, jint root)
{
   jint *rcount;
   jint *dps;
   jboolean isCopy=JNI_TRUE;
   jbyte * sendptr=getBufPtr(env,sendbuf,sendtype);
   jbyte * recvptr=getBufPtr(env,recvbuf,recvtype);
   rcount=(*env)->GetIntArrayElements(env,recvcount,&isCopy);
   dps=(*env)->GetIntArrayElements(env,displs,&isCopy);
   MPI_Gatherv(sendptr+sendoffset,sendcount,
	(MPI_Datatype)((*env)->GetLongField(env,sendtype,DatatypehandleID)),
	recvptr+recvoffset,(int*) rcount,
	(int*)dps,(MPI_Datatype)((*env)->GetLongField(env,recvtype,DatatypehandleID)),
	root,(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   (*env)->ReleaseIntArrayElements(env,recvcount,rcount,0);
   (*env)->ReleaseIntArrayElements(env,displs,dps,0);
   releaseBufPtr(env,sendbuf,sendtype,sendptr);
   releaseBufPtr(env,recvbuf,recvtype,recvptr);
}

/*
 * Class:     mpi_Intracomm
 * Method:    Scatter
 * Signature:
(Ljava/lang/Object;IILmpi/Datatype;Ljava/lang/Object;IILmpi/Dataty
pe;I)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_scatter
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, jint
sendcount, jobject sendtype, jobject recvbuf, jint recvoffset, jint recvcount,
jobject recvtype, jint root)
{
   jbyte * sendptr=getBufPtr(env,sendbuf,sendtype);
   jbyte * recvptr=getBufPtr(env,recvbuf,recvtype);
   MPI_Scatter(sendptr+sendoffset, sendcount,
	(MPI_Datatype)((*env)->GetLongField(env,sendtype,DatatypehandleID)), 
	recvptr+recvoffset, recvcount,
	(MPI_Datatype)((*env)->GetLongField(env,recvtype,DatatypehandleID)),
	root,(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   releaseBufPtr(env,sendbuf,sendtype,sendptr);
   releaseBufPtr(env,recvbuf,recvtype,recvptr);
}


/*
 * Class:     mpi_Intracomm
 * Method:    Scatterv
 * Signature:
(Ljava/lang/Object;II[ILmpi/Datatype;Ljava/lang/Object;I[ILmpi/Dat
atype;I)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_scatterv
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, 
	jintArray sendcount,jintArray displs, 
	jobject sendtype, jobject recvbuf, jint recvoffset,
	jint recvcount,jobject recvtype, jint root)
{
   jint *scount;
   jint *dps;
   jboolean isCopy=JNI_TRUE;
   jbyte * sendptr=getBufPtr(env,sendbuf,sendtype);
   jbyte * recvptr=getBufPtr(env,recvbuf,recvtype);
   scount=(*env)->GetIntArrayElements(env,sendcount,&isCopy);
   dps=(*env)->GetIntArrayElements(env,displs,&isCopy);
   MPI_Scatterv(sendptr+sendoffset,(int*)scount,(int*) dps,
	(MPI_Datatype)((*env)->GetLongField(env,sendtype,DatatypehandleID)),
	recvptr+recvoffset, recvcount,
	(MPI_Datatype)((*env)->GetLongField(env,recvtype,DatatypehandleID)),
	root,(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   (*env)->ReleaseIntArrayElements(env,sendcount,scount,0);
   (*env)->ReleaseIntArrayElements(env,displs,dps,0);
   releaseBufPtr(env,sendbuf,sendtype,sendptr);
   releaseBufPtr(env,recvbuf,recvtype,recvptr);
}

/*
 * Class:     mpi_Intracomm
 * Method:    Allgather
 * Signature:
(Ljava/lang/Object;IILmpi/Datatype;Ljava/lang/Object;IILmpi/Dataty
pe;)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_allgather
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, jint
sendcount, jobject sendtype, jobject recvbuf, jint recvoffset, jint recvcount,
jobject recvtype)
{
   jbyte * sendptr=getBufPtr(env,sendbuf,sendtype);
   jbyte * recvptr=getBufPtr(env,recvbuf,recvtype);
   MPI_Allgather(sendptr+sendoffset, sendcount,
	(MPI_Datatype)((*env)->GetLongField(env,sendtype,DatatypehandleID)), 
	recvptr+recvoffset, recvcount,
	(MPI_Datatype)((*env)->GetLongField(env,recvtype,DatatypehandleID)),
	(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   releaseBufPtr(env,sendbuf,sendtype,sendptr);
   releaseBufPtr(env,recvbuf,recvtype,recvptr);
}

/*
 * Class:     mpi_Intracomm
 * Method:    Allgatherv
 * Signature:
(Ljava/lang/Object;IILmpi/Datatype;Ljava/lang/Object;I[I[ILmpi/Dat
atype;)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_allgatherv
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, jint sendcount,jobject sendtype, jobject recvbuf, jint recvoffset,
jintArray recvcount, jintArray displs, jobject recvtype)
{
   jint *rcount;
   jint *dps;
   jboolean isCopy=JNI_TRUE;
   jbyte * sendptr=getBufPtr(env,sendbuf,sendtype);
   jbyte * recvptr=getBufPtr(env,recvbuf,recvtype);
   rcount=(*env)->GetIntArrayElements(env,recvcount,&isCopy);
   dps=(*env)->GetIntArrayElements(env,displs,&isCopy);
   MPI_Allgatherv(sendptr+sendoffset,sendcount,
	(MPI_Datatype)((*env)->GetLongField(env,sendtype,DatatypehandleID)),
	recvptr+recvoffset,(int*) rcount,
	(int*)dps,(MPI_Datatype)((*env)->GetLongField(env,recvtype,DatatypehandleID)),
	(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   (*env)->ReleaseIntArrayElements(env,recvcount,rcount,0);
   (*env)->ReleaseIntArrayElements(env,displs,dps,0);
   releaseBufPtr(env,sendbuf,sendtype,sendptr);
   releaseBufPtr(env,recvbuf,recvtype,recvptr);
}
 
/*
 * Class:     mpi_Intracomm
 * Method:    Alltoall
 * Signature:
(Ljava/lang/Object;IILmpi/Datatype;Ljava/lang/Object;IILmpi/Dataty
pe;)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_alltoall
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, jint
sendcount, jobject sendtype, jobject recvbuf, jint recvoffset, jint recvcount,
jobject recvtype)
{
   jbyte * sendptr=getBufPtr(env,sendbuf,sendtype);
   jbyte * recvptr=getBufPtr(env,recvbuf,recvtype);
   MPI_Alltoall(sendptr+sendoffset, sendcount,
	(MPI_Datatype)((*env)->GetLongField(env,sendtype,DatatypehandleID)), 
	recvptr+recvoffset, recvcount,
	(MPI_Datatype)((*env)->GetLongField(env,recvtype,DatatypehandleID)),
	(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   releaseBufPtr(env,sendbuf,sendtype,sendptr);
   releaseBufPtr(env,recvbuf,recvtype,recvptr);
}

/*
 * Class:     mpi_Intracomm
 * Method:    Alltoallv
 * Signature:
(Ljava/lang/Object;II[ILmpi/Datatype;Ljava/lang/Object;I[I[ILmpi/D
atatype;)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_alltoallv
   (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, jintArray sendcount, 
	  jintArray sdispls, jobject sendtype, jobject recvbuf, jint recvoffset,
	  jintArray recvcount, jintArray rdispls, jobject recvtype)
{
   jint *rcount, *scount;
   jint *sdps,*rdps;
   jboolean isCopy=JNI_TRUE;
   jbyte * sendptr=getBufPtr(env,sendbuf,sendtype);
   jbyte * recvptr=getBufPtr(env,recvbuf,recvtype);
   scount=(*env)->GetIntArrayElements(env,sendcount,&isCopy);
   rcount=(*env)->GetIntArrayElements(env,recvcount,&isCopy);
   sdps=(*env)->GetIntArrayElements(env,sdispls,&isCopy);
   rdps=(*env)->GetIntArrayElements(env,rdispls,&isCopy);
   MPI_Alltoallv(sendptr+sendoffset,(int*)scount, (int*)sdps,
	(MPI_Datatype)((*env)->GetLongField(env,sendtype,DatatypehandleID)),
	recvptr+recvoffset, (int*)rcount,
	(int*)rdps,(MPI_Datatype)((*env)->GetLongField(env,recvtype,DatatypehandleID)),
	(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   (*env)->ReleaseIntArrayElements(env,recvcount,rcount,0);
   (*env)->ReleaseIntArrayElements(env,sendcount,scount,0);
   (*env)->ReleaseIntArrayElements(env,sdispls,sdps,0);
   (*env)->ReleaseIntArrayElements(env,rdispls,rdps,0);
   releaseBufPtr(env,sendbuf,sendtype,sendptr);
   releaseBufPtr(env,recvbuf,recvtype,recvptr);
}
 
/*
 * Class:     mpi_Intracomm
 * Method:    Reduce
 * Signature:
(Ljava/lang/Object;ILjava/lang/Object;IILmpi/Datatype;Lmpi/Op;I)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_reduce
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, 
	  jobject recvbuf, jint recvoffset, jint count,
	  jobject type, jobject op, jint root)
{
   jbyte * sendptr=getBufPtr(env,sendbuf,type);
   jbyte * recvptr=getBufPtr(env,recvbuf,type);
   MPI_Reduce(sendptr+sendoffset,
		recvptr+recvoffset, count,
	(MPI_Datatype)((*env)->GetLongField(env,type,DatatypehandleID)),
	(MPI_Op)((*env)->GetLongField(env,op,OphandleID)),
	root,(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   releaseBufPtr(env,sendbuf,type,sendptr);
   releaseBufPtr(env,recvbuf,type,recvptr);
}

/*
 * Class:     mpi_Intracomm
 * Method:    Allreduce
 * Signature:
(Ljava/lang/Object;ILjava/lang/Object;IILmpi/Datatype;Lmpi/Op;)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_allreduce
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, 
	  jobject recvbuf, jint recvoffset, jint count,
	  jobject type, jobject op)
{
   jbyte * sendptr=getBufPtr(env,sendbuf,type);
   jbyte * recvptr=getBufPtr(env,recvbuf,type);
   MPI_Allreduce(sendptr+sendoffset,
		recvptr+recvoffset, count,
	(MPI_Datatype)((*env)->GetLongField(env,type,DatatypehandleID)),
	(MPI_Op)((*env)->GetLongField(env,op,OphandleID)),
	(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   releaseBufPtr(env,sendbuf,type,sendptr);
   releaseBufPtr(env,recvbuf,type,recvptr);
}

/*
 * Class:     mpi_Intracomm
 * Method:    Reduce_scatter
 * Signature:
(Ljava/lang/Object;ILjava/lang/Object;I[ILmpi/Datatype;Lmpi/Op;)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_reduce_1scatter
  (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset,
	  jobject recvbuf, jint recvoffset, jintArray recvcount,
	  jobject type, jobject op)
{
   jint *rcount;
   jboolean isCopy=JNI_TRUE;
   jbyte * sendptr=getBufPtr(env,sendbuf,type);
   jbyte * recvptr=getBufPtr(env,recvbuf,type);
   rcount=(*env)->GetIntArrayElements(env,recvcount,&isCopy);
   MPI_Reduce_scatter(sendptr+sendoffset,
		recvptr+recvoffset,(int*) rcount,
	(MPI_Datatype)((*env)->GetLongField(env,type,DatatypehandleID)),
	(MPI_Op)((*env)->GetLongField(env,op,OphandleID)),
	(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   (*env)->ReleaseIntArrayElements(env,recvcount,rcount,0);
   releaseBufPtr(env,sendbuf,type,sendptr);
   releaseBufPtr(env,recvbuf,type,recvptr);
}

/*
 * Class:     mpi_Intracomm
 * Method:    Scan
 * Signature:
(Ljava/lang/Object;ILjava/lang/Object;IILmpi/Datatype;Lmpi/Op;)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_scan
 (JNIEnv *env, jobject jthis, jobject sendbuf, jint sendoffset, 
	  jobject recvbuf, jint recvoffset, jint count,
	 jobject type, jobject op)
{
   jbyte * sendptr=getBufPtr(env,sendbuf,type);
   jbyte * recvptr=getBufPtr(env,recvbuf,type);
   MPI_Scan(sendptr+sendoffset,
		recvptr+recvoffset, count,
	(MPI_Datatype)((*env)->GetLongField(env,type,DatatypehandleID)),
	(MPI_Op)((*env)->GetLongField(env,op,OphandleID)),
	(MPI_Comm)((*env)->GetLongField(env,jthis,CommhandleID)));
   releaseBufPtr(env,sendbuf,type,sendptr);
   releaseBufPtr(env,recvbuf,type,recvptr);
}

