#include <mpi.h>
#include "mpi_Op.h"

jfieldID OphandleID;

jmethodID callID;
jclass datatype_class, object_class;
jmethodID DatatypehandleConstructorID;

#define MAX_OP_NUM 8
int gOpIndex;
int validOp[MAX_OP_NUM];  /* 0: empty, may be alloc; 1: occupied */
JNIEnv* gjenv[MAX_OP_NUM];
jobject gOpfunction[MAX_OP_NUM];
MPI_Op gOp[MAX_OP_NUM];

/* Java_call_fn, are for function pointer in MPI_Op_create.
   Refer to P19 in document MPI_Java Interface.
   xli 1/29/98
*/
#define JAVA_CALL_FUN(NUM) \
	void  Java_call_fn_##NUM\
	(void *inve,void *inoutvec,int *len,MPI_Datatype *type) \
{   jobject jinvec=(*(gjenv[NUM]))->AllocObject(gjenv[NUM],object_class); \
    jobject joutvec=(*(gjenv[NUM]))->AllocObject(gjenv[NUM],object_class); \
    jobject jtype=(*(gjenv[NUM]))->NewObject(gjenv[NUM],datatype_class, \
        DatatypehandleConstructorID, type); \
    (*(gjenv[NUM]))->CallVoidMethod(gjenv[NUM],(jobject)gOpfunction[NUM], \
            callID,jinvec,joutvec,jtype); } 

JAVA_CALL_FUN(0)
JAVA_CALL_FUN(1)
JAVA_CALL_FUN(2)
JAVA_CALL_FUN(3)
JAVA_CALL_FUN(4)
JAVA_CALL_FUN(5)
JAVA_CALL_FUN(6)
JAVA_CALL_FUN(7)

typedef void (*PF)(void *,void *,int *, MPI_Datatype *);
PF Java_call_fn[]={&Java_call_fn_0, &Java_call_fn_1,
	&Java_call_fn_2, &Java_call_fn_3,
	&Java_call_fn_4, &Java_call_fn_5,
	&Java_call_fn_6, &Java_call_fn_7,
	};	
/*
 * Class:     mpi_Op
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Op_init
  (JNIEnv *env, jclass thisClass) 
{
  int i;
  OphandleID=(*env)->GetFieldID(env,thisClass,"handle","J");
  for(i=0;i<MAX_OP_NUM;i++)
    validOp[i]=0;
}

/*
 * Class:     mpi_Op
 * Method:    GetOp
 * Signature: (I)J
 */
JNIEXPORT void JNICALL Java_mpi_Op_GetOp
  (JNIEnv *env, jobject jthis, jint type) 
{
  static MPI_Op Ops[] = { MPI_OP_NULL, MPI_MAX, MPI_MIN, MPI_SUM,
    MPI_PROD, MPI_LAND, MPI_BAND, MPI_LOR, MPI_BOR, MPI_LXOR,
    MPI_BXOR, MPI_MINLOC, MPI_MAXLOC};
  (*env)->SetLongField(env,jthis, OphandleID, (jlong)Ops[type]);
}

/* New or Delete Operand, jenv and function from the pool of maxsize 8.
   xli 2/3/97 */

int newOp(jobject function)
{
  int i=0;
  while(i<MAX_OP_NUM && validOp[i]!=0)
    i++;
  if(i==MAX_OP_NUM)
  {  /* printf("User_defined Operands are 8. Cann't create any more.\n"); */
     return -1;
  }
  validOp[i]=1;
  gOpfunction[i]=function;
  return i;
}

void deleteOp(MPI_Op op)
{
  int i;
  while(i<MAX_OP_NUM && gOp[i]!=op)
    i++;
  if(i==MAX_OP_NUM) return;
  validOp[i]=0;
}

/*
 * Class:     mpi_Op
 * Method:    GetOp
 * Signature: (Lmpi/User_function;Z)V
 */
JNIEXPORT void JNICALL Java_mpi_Op_GetOp__Lmpi_User_1function_2Z
  (JNIEnv *env, jobject jthis, jobject function, jboolean commute)
{
  jclass userfunction_class=(*env)->FindClass(env,"mpi/User_function");
  jmethodID callID=(*env)->GetMethodID(env,userfunction_class,
 	"call","(Ljava/lang/Object;Ljava/lang/Object;Lmpi/Datatype;)V");

  object_class=(*env)->FindClass(env,"Object");

  datatype_class=(*env)->FindClass(env,"mpi/Datatype");
  DatatypehandleConstructorID = (*env)->GetMethodID(env,
     datatype_class, "<init>", "(I)V");
  
  if((gOpIndex=newOp(function))==-1)
  {  /* printf("Failure in Creating Op\n"); */
     return;
  }
  gjenv[gOpIndex]=env;

  MPI_Op_create(Java_call_fn[gOpIndex],commute, &(gOp[gOpIndex]));
  
  (*env)->SetLongField(env,jthis, OphandleID, (jlong)(gOp[gOpIndex]));
}

/*
 * Class:     mpi_Op
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Op_finalize
  (JNIEnv *env, jobject jthis)
{
  MPI_Op op;
  op=(MPI_Op)((*env)->GetLongField(env,jthis,OphandleID));
  MPI_Op_free(&op);
  deleteOp(op);
  (*env)->SetLongField(env,jthis, OphandleID,0);
}
